"""
Status enums for edx-proctoring
"""

from edx_proctoring.exceptions import ProctoredExamBadReviewStatus


class ProctoredExamStudentAttemptStatus:
    """
    A class to enumerate the various status that an attempt can have

    IMPORTANT: Since these values are stored in a database, they are system
    constants and should not be language translated, since translations
    might change over time.
    """

    # the student is eligible to decide if he/she wants to pursue credit
    eligible = 'eligible'

    # the attempt record has been created, but the exam has not yet
    # been started
    created = 'created'

    # the student has clicked on the external
    # software download link
    download_software_clicked = 'download_software_clicked'

    # the attempt is ready to start but requires
    # user to acknowledge that he/she wants to start the exam
    ready_to_start = 'ready_to_start'

    # the student has started the exam and is
    # in the process of completing the exam
    started = 'started'

    # the student has completed the exam
    ready_to_submit = 'ready_to_submit'

    #
    # The follow statuses below are considered in a 'completed' state
    # and we will not allow transitions to status above this mark
    #

    # the student declined to take the exam as a proctored exam
    declined = 'declined'

    # the exam has timed out
    timed_out = 'timed_out'

    # the student has submitted the exam for proctoring review
    submitted = 'submitted'

    # the student has submitted the exam for proctoring review
    second_review_required = 'second_review_required'

    # the exam has been verified and approved
    verified = 'verified'

    # the exam has been rejected
    rejected = 'rejected'

    # the exam is believed to be in error
    error = 'error'

    # the course end date has passed
    expired = 'expired'

    # the onboarding attempt has been reset
    onboarding_reset = 'onboarding_reset'

    # onboarding failure states
    # the user hasn't taken an onboarding exam
    onboarding_missing = 'onboarding_missing'
    # the onboarding exam is pending review
    onboarding_pending = 'onboarding_pending'
    # the user failed onboarding
    onboarding_failed = 'onboarding_failed'
    # the onboarding data expired
    onboarding_expired = 'onboarding_expired'

    onboarding_errors = (onboarding_missing, onboarding_pending, onboarding_failed, onboarding_expired)

    @classmethod
    def is_completed_status(cls, status):
        """
        Returns a boolean if the passed in status is in a "completed" state, meaning
        that it cannot go backwards in state
        """
        return status in [
            cls.declined, cls.timed_out, cls.submitted, cls.second_review_required,
            cls.verified, cls.rejected, cls.error, cls.onboarding_missing, cls.onboarding_pending,
            cls.onboarding_failed, cls.onboarding_expired
        ]

    @classmethod
    def is_incomplete_status(cls, status):
        """
        Returns a boolean if the passed in status is in an "incomplete" state.
        """
        return status in [
            cls.eligible, cls.created, cls.download_software_clicked, cls.ready_to_start, cls.started,
            cls.ready_to_submit
        ]

    @classmethod
    def needs_credit_status_update(cls, to_status):
        """
        Returns a boolean if the passed in to_status calls for an update to the credit requirement status.
        """
        return to_status in [
            cls.verified, cls.rejected, cls.declined, cls.submitted, cls.error
        ]

    @classmethod
    def needs_grade_override(cls, to_status):
        """
        Returns a boolean if the passed in to_status calls for an override of the learner's grade.
        """
        return to_status in [
            cls.rejected
        ]

    @classmethod
    def is_a_cascadable_failure(cls, to_status):
        """
        Returns a boolean if the passed in to_status has a failure that needs to be cascaded
        to other unattempted exams.
        """
        return to_status in [
            cls.declined
        ]

    @classmethod
    def is_valid_status(cls, status):
        """
        Makes sure that passed in status string is valid
        """
        return cls.is_completed_status(status) or cls.is_incomplete_status(status)

    @classmethod
    def is_pre_started_status(cls, status):
        """
        Returns a boolean if the status passed is prior to "started" state.
        """
        return status in [
            cls.created, cls.download_software_clicked, cls.ready_to_start
        ]

    @classmethod
    def is_in_progress_status(cls, status):
        """
        Returns a boolean if the status passed is "in progress".
        """
        return status in [
            cls.started, cls.ready_to_submit
        ]


class ReviewStatus:
    """
    Standard review statuses
    """
    passed = 'passed'
    violation = 'violation'
    suspicious = 'suspicious'
    not_reviewed = 'not_reviewed'

    @classmethod
    def validate(cls, status):
        """
        Validate review status
        """
        if status not in (cls.passed, cls.violation, cls.suspicious, cls.not_reviewed):
            raise ProctoredExamBadReviewStatus(status)
        return True


class SoftwareSecureReviewStatus:
    """
    These are the valid review statuses from
    SoftwareSecure
    """

    clean = 'Clean'
    violation = 'Rules Violation'
    suspicious = 'Suspicious'
    not_reviewed = 'Not Reviewed'

    passing_statuses = [
        clean,
        violation]
    failing_statuses = [
        not_reviewed,
        suspicious]
    notify_support_for_status = suspicious

    from_standard_status = {
        ReviewStatus.passed: clean,
        ReviewStatus.violation: violation,
        ReviewStatus.suspicious: suspicious,
        ReviewStatus.not_reviewed: not_reviewed,
    }

    to_standard_status = {v: k for k, v in from_standard_status.items()}

    @classmethod
    def validate(cls, status):
        """
        Validates the status, or raises ProctoredExamBadReviewStatus
        """
        if status not in cls.passing_statuses + cls.failing_statuses:
            err_msg = (
                'Received unexpected reviewStatus field value from payload. '
                f'Was {status}.'
            )
            raise ProctoredExamBadReviewStatus(err_msg)
        return True


class InstructorDashboardOnboardingAttemptStatus:
    """
    A class to enumerate the different statuses a proctored exam attempt
    in an onboarding exam can have and to map them to internal database statutes.
    These are intended to be used in externally facing applications, such
    as the Instructor Dashboard.
    """
    not_started = 'not_started'
    setup_started = 'setup_started'
    onboarding_started = 'onboarding_started'
    submitted = 'submitted'
    rejected = 'rejected'
    verified = 'verified'
    error = 'error'
    expired = 'expired'

    # The following status is not a true attempt status, but is used when the
    # user's onboarding profile is approved in a different course.
    other_course_approved = 'other_course_approved'

    # The following status is not a true attempt status that has a corresponding database
    # state. This is a consequence of a bug in our software that allows a learner to end up
    # with their only or their most recent exam attempt being in the "onboarding_reset" state.
    # The learner should not end up in this state, but while we work on a fix, we should not
    # display "null" in the Instructor Dashboard Student Onboarding Panel.
    # TODO: remove as part of MST-745
    onboarding_reset_past_due = 'onboarding_reset_past_due'

    onboarding_statuses = {
        ProctoredExamStudentAttemptStatus.created: setup_started,
        ProctoredExamStudentAttemptStatus.download_software_clicked: setup_started,
        ProctoredExamStudentAttemptStatus.ready_to_start: setup_started,
        ProctoredExamStudentAttemptStatus.started: onboarding_started,
        ProctoredExamStudentAttemptStatus.ready_to_submit: onboarding_started,
        ProctoredExamStudentAttemptStatus.submitted: submitted,
        ProctoredExamStudentAttemptStatus.rejected: rejected,
        ProctoredExamStudentAttemptStatus.verified: verified,
        ProctoredExamStudentAttemptStatus.error: error,
        ProctoredExamStudentAttemptStatus.expired: expired
    }

    @classmethod
    def get_onboarding_status_from_attempt_status(cls, status):
        """
        Get the externally facing proctored exam attempt onboarding status
        from an internal database proctored exam attempt status.

        Parameters:
            * status: a ProctoredExamStudentAttemptStatus status
        """
        if status:
            return cls.onboarding_statuses.get(status)
        return cls.not_started


class VerificientOnboardingProfileStatus:
    """
    A class for mapping onboarding statuses from Verificient's onboarding status endpoint to
    internal database values.
    """
    no_profile = 'no-profile'
    approved = 'approved-in-course'
    other_course_approved = 'approved-in-different-course'
    rejected = 'rejected'
    expired = 'expired'
    pending = 'pending'

    profile_status_mapping = {
        no_profile: None,
        approved: InstructorDashboardOnboardingAttemptStatus.verified,
        other_course_approved: InstructorDashboardOnboardingAttemptStatus.other_course_approved,
        rejected: InstructorDashboardOnboardingAttemptStatus.rejected,
        expired: InstructorDashboardOnboardingAttemptStatus.expired,
        pending: InstructorDashboardOnboardingAttemptStatus.submitted
    }

    # reverse map: leave out the none and replace with not_started for filter use
    filter_status_mapping = {v: k for k, v in profile_status_mapping.items() if v}
    filter_status_mapping[InstructorDashboardOnboardingAttemptStatus.not_started] = no_profile

    @classmethod
    def get_edx_status_from_profile_status(cls, api_status):
        """
        Get the internal attempt status given a status from the onboarding api

        Parameters:
            * status (str): status from Verficient's onboarding API endpoint
        """
        return cls.profile_status_mapping.get(api_status)

    @classmethod
    def get_profile_status_from_filter(cls, filter_status):
        """
        Get the verificient profile status given an edx status for filtering

        Parameters:
            * status (str): edX onboarding status
        """
        return cls.filter_status_mapping.get(filter_status)

    @classmethod
    def get_instructor_status_from_profile_status(cls, api_status):
        """
        Get the instructor onboarding status given a status from the onboarding api

        Parameters:
            * status (str): status from Verficient's onboarding API endpoint
        """
        if api_status == cls.no_profile:
            return InstructorDashboardOnboardingAttemptStatus.not_started
        return cls.get_edx_status_from_profile_status(api_status)



"""Django Rules for edx_proctoring"""

import rules


@rules.predicate
def is_in_reviewer_group(user, attempt):
    """
    Returns whether user is in a group allowed to review this attempt
    """
    backend_group = f"{attempt['proctored_exam']['backend']}_review"
    return user.groups.filter(name=backend_group).exists()


# pylint: disable-next=unsupported-binary-operation
rules.add_perm('edx_proctoring.can_review_attempt', is_in_reviewer_group | rules.is_staff)



"""
Tests for the reset_attempts management command
"""
from tempfile import NamedTemporaryFile

import ddt

from django.core.management import call_command

from edx_proctoring.api import create_exam
from edx_proctoring.models import ProctoredExamStudentAttempt
from edx_proctoring.statuses import ProctoredExamStudentAttemptStatus
from edx_proctoring.tests.utils import LoggedInTestCase


@ddt.ddt
class ResetAttemptsTests(LoggedInTestCase):
    """
    Coverage of the reset_attempts.py file
    """

    def setUp(self):
        """
        Build up test data
        """
        super().setUp()
        self.exam_id = create_exam(
            course_id='a/b/c',
            content_id='bar',
            exam_name='Test Exam',
            time_limit_mins=90
        )

        self.num_attempts = 10

        user_list = self.create_batch_users(self.num_attempts)
        for user in user_list:
            ProctoredExamStudentAttempt.objects.create(
                proctored_exam_id=self.exam_id,
                user_id=user.id,
                external_id='foo',
                status=ProctoredExamStudentAttemptStatus.created,
                allowed_time_limit_mins=10,
                taking_as_proctored=True,
                is_sample_attempt=False
            )

    @ddt.data(
        5,
        7,
        10,
    )
    def test_run_command(self, num_to_delete):
        """
        Run the management command
        """
        ids = list(ProctoredExamStudentAttempt.objects.all().values_list('id', flat=True))[:num_to_delete]

        with NamedTemporaryFile() as file:
            with open(file.name, 'w') as writing_file:
                for num in ids:
                    writing_file.write(str(num) + '\n')

            call_command(
                'reset_attempts',
                batch_size=2,
                sleep_time=0,
                file_path=file.name,
            )

        attempts = ProctoredExamStudentAttempt.objects.all()
        self.assertEqual(len(attempts), self.num_attempts - num_to_delete)



# -*- coding: utf-8 -*-
"""
edx_proctoring Django application initialization.
"""

import json
import os
import os.path
import warnings

from stevedore.extension import ExtensionManager

from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

BACKEND_CONFIGURATION_ALLOW_LIST = [
    'base_url',
    'client_id',
    'client_secret',
    'crypto_key',
    'default_rules',
    'exam_register_endpoint',
    'exam_sponsor',
    'has_dashboard',
    'help_center_article_url',
    'integration_specific_email',
    'learner_notification_from_email',
    'needs_oauth',
    'organization',
    'passing_statuses',
    'ping_interval',
    'secret_key',
    'secret_key_id',
    'send_email',
    'software_download_url',
    'supports_onboarding',
    'tech_support_email',
    'tech_support_phone',
    'token_expiration_time',
    'verbose_name',
    'video_review_aes_key',
]


def make_worker_config(backends, out='/tmp/workers.json'):
    """
    Generates a config json file used for edx-platform's webpack.common.config.js
    """
    if not getattr(settings, 'NODE_MODULES_ROOT', None):
        return False
    config = {}
    for backend in backends:
        try:
            package = backend.npm_module
            package_file = os.path.join(settings.NODE_MODULES_ROOT, package, 'package.json')
            with open(package_file, 'r', encoding='utf-8') as package_fp:
                package_json = json.load(package_fp)
            main_file = package_json['main']
            config[package] = ['babel-polyfill', os.path.join(settings.NODE_MODULES_ROOT, package, main_file)]
        except AttributeError:
            # no npm module defined
            continue
        except IOError:
            warnings.warn(f'Proctoring backend {backend.__class__} defined an npm module,'
                          f'but it is not installed at {package_file!r}')
        except KeyError:
            warnings.warn(f'{package_file!r} does not contain a `main` entry')
    if config:
        try:
            with open(out, 'wb+') as outfp:
                outfp.write(json.dumps(config).encode('utf-8'))
        except IOError:
            warnings.warn(f"Could not write worker config to {out}")
        else:
            # make sure that this file is group writable, because it may be written by different users
            os.chmod(out, 0o664)
            return True
    return False


class EdxProctoringConfig(AppConfig):
    """
    Configuration for the edx_proctoring Django application.
    """

    name = 'edx_proctoring'
    plugin_app = {
        'url_config': {
            'lms.djangoapp': {
                'namespace': 'edx_proctoring',
                'regex': '^api/',
                'relative_path': 'urls',
            },
            'cms.djangoapp': {
                'namespace': 'edx_proctoring',
                'regex': '^api/',
                'relative_path': 'instructor_dashboard_exam_urls',
            },
        },
        'settings_config': {
            'lms.djangoapp': {
                'common': {'relative_path': 'settings.common'},
                'production': {'relative_path': 'settings.production'},
            },
            'cms.djangoapp': {
                'common': {'relative_path': 'settings.common'},
                'production': {'relative_path': 'settings.production'},
            }

        },
    }

    def get_backend_choices(self):
        """
        Returns an iterator of available backends:
        backend_name, verbose name
        """
        for name, backend in self.backends.items():
            yield name, getattr(backend, 'verbose_name', 'Unknown')

    def get_backend(self, name=None):
        """
        Returns an instance of the proctoring backend.

        :param str name: Name of entrypoint in openedx.proctoring
        :param dict options: Keyword arguments to use when instantiating the backend
        """
        if name is None:
            try:
                name = settings.PROCTORING_BACKENDS['DEFAULT']
            except (KeyError, AttributeError) as exc:
                raise ImproperlyConfigured("No default proctoring backend set in settings.PROCTORING_BACKENDS") \
                    from exc
        try:
            return self.backends[name]
        except KeyError as error:
            raise NotImplementedError(f"No proctoring backend configured for '{name}'.  "
                                      f"Available: {list(self.backends)}") from error

    def ready(self):
        """
        Loads the available proctoring backends
        """
        # pylint: disable=unused-import
        # pylint: disable=import-outside-toplevel
        from edx_proctoring import handlers, signals
        config = settings.PROCTORING_BACKENDS

        self.backends = {}  # pylint: disable=W0201
        for extension in ExtensionManager(namespace='openedx.proctoring'):
            name = extension.name
            try:
                options = {
                    key: val for (key, val) in config[name].items()
                    if key in BACKEND_CONFIGURATION_ALLOW_LIST
                }
                self.backends[name] = extension.plugin(**options)
            except KeyError:
                pass
        make_worker_config(list(self.backends.values()), out=os.path.join(settings.ENV_ROOT, 'workers.json'))



# Generated by Django 3.2.20 on 2023-07-27 21:12

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('edx_proctoring', '0024_delete_proctoredexamstudentattempthistory'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='historicalproctoredexam',
            options={'get_latest_by': ('history_date', 'history_id'), 'ordering': ('-history_date', '-history_id'), 'verbose_name': 'historical proctored exam', 'verbose_name_plural': 'historical proctored exams'},
        ),
        migrations.AlterModelOptions(
            name='historicalproctoredexamstudentattempt',
            options={'get_latest_by': ('history_date', 'history_id'), 'ordering': ('-history_date', '-history_id'), 'verbose_name': 'historical proctored exam attempt', 'verbose_name_plural': 'historical proctored exam attempts'},
        ),
    ]



"""
Django management command to re-trigger the status update of a ProctoredExamStudentAttempt
for a specific proctored exam.
"""

import logging
import time

from django.core.management.base import BaseCommand

from edx_proctoring.api import update_attempt_status
from edx_proctoring.models import ProctoredExamStudentAttempt

log = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Django Management command to update is_attempt_active field on review models
    """

    def add_arguments(self, parser):
        parser.add_argument(
            '--batch_size',
            action='store',
            dest='batch_size',
            type=int,
            default=300,
            help='Maximum number of attempts to process. '
                 'This helps avoid overloading the database while updating large amount of data.'
        )
        parser.add_argument(
            '--sleep_time',
            action='store',
            dest='sleep_time',
            type=int,
            default=10,
            help='Sleep time in seconds between update of batches'
        )

        parser.add_argument(
            '--exam_id',
            action='store',
            dest='exam_id',
            type=int,
            help='Exam ID to process attempts for.'
        )

    def handle(self, *args, **options):
        """
        Management command entry point, simply call into the signal firing
        """

        batch_size = options['batch_size']
        sleep_time = options['sleep_time']
        exam_id = options['exam_id']

        # get all attempts for specific exam id
        exam_attempts = ProctoredExamStudentAttempt.objects.filter(proctored_exam_id=exam_id)

        attempt_count = 0

        # for each of those attempts, get id and status
        for attempt in exam_attempts:
            current_status = attempt.status
            current_id = attempt.id

            log.info(
                'Triggering attempt status update for attempt_id=%(attempt_id)s with status=%(status)s',
                {'attempt_id': current_id, 'status': current_status}
            )
            # need to use update_attempt_status because this function will trigger grade + credit updates
            update_attempt_status(current_id, current_status)
            attempt_count += 1

            if attempt_count == batch_size:
                attempt_count = 0
                time.sleep(sleep_time)



"""
Tests for the update_attempts_for_exam management command
"""

from mock import patch

from django.contrib.auth import get_user_model
from django.core.management import call_command

from edx_proctoring.api import create_exam, create_exam_attempt, update_attempt_status
from edx_proctoring.models import ProctoredExamStudentAttempt
from edx_proctoring.runtime import set_runtime_service
from edx_proctoring.statuses import ProctoredExamStudentAttemptStatus
from edx_proctoring.tests.test_services import MockCertificateService, MockCreditService, MockGradesService
from edx_proctoring.tests.utils import LoggedInTestCase

User = get_user_model()


class TestUpdateAttemptsForExam(LoggedInTestCase):
    """
    Coverage of the update_attempts_for_exam.py file
    """

    def setUp(self):
        """
        Build up test data
        """
        super().setUp()
        set_runtime_service('credit', MockCreditService())
        set_runtime_service('grades', MockGradesService())
        set_runtime_service('certificates', MockCertificateService())

    def test_run_command(self):
        """
        Run the management command
        """
        exam_id = create_exam(
            course_id='a/b/c',
            content_id='bar',
            exam_name='Test Exam 1',
            time_limit_mins=90
        )

        # create three users and three exam attempts
        for i in range(3):
            other_user = User.objects.create(username='otheruser'+str(i), password='test')
            attempt_id = create_exam_attempt(exam_id, other_user.id, taking_as_proctored=True)
            update_attempt_status(attempt_id, ProctoredExamStudentAttemptStatus.verified)

        with patch.object(MockCreditService, 'set_credit_requirement_status') as mock_credit:
            call_command(
                'update_attempts_for_exam',
                batch_size=2,
                sleep_time=0,
                exam_id=exam_id
            )
            mock_credit.assert_called()

        # make sure status stays the same
        attempts = ProctoredExamStudentAttempt.objects.filter(status=ProctoredExamStudentAttemptStatus.verified)
        self.assertEqual(len(attempts), 3)



"""
Tests for the MFE proctored exam views.
"""
import json
from itertools import product
from urllib.parse import urlencode

import ddt
from mock import patch
from opaque_keys.edx.locator import BlockUsageLocator

from django.conf import settings
from django.contrib.auth import get_user_model
from django.test.utils import override_settings
from django.urls import reverse
from django.utils import timezone

from edx_proctoring.api import get_exam_by_id, get_review_policy_by_exam_id
from edx_proctoring.exceptions import BackendProviderNotConfigured, ProctoredExamNotFoundException
from edx_proctoring.models import ProctoredExam, ProctoredExamStudentAllowance
from edx_proctoring.runtime import set_runtime_service
from edx_proctoring.statuses import ProctoredExamStudentAttemptStatus
from edx_proctoring.utils import humanized_time

from .test_services import MockLearningSequencesService, MockScheduleItemData
from .utils import ProctoredExamTestCase

User = get_user_model()


@override_settings(LEARNING_MICROFRONTEND_URL='https//learningmfe', ACCOUNT_MICROFRONTEND_URL='https//localhost')
@ddt.ddt
class ProctoredExamAttemptsMFEViewTests(ProctoredExamTestCase):
    """
    Tests for the ProctoredExamView called from MFE application.
    """
    def setUp(self):
        """
        Initialize
        """
        super().setUp()
        self.timed_exam_id = self._create_timed_exam()
        self.proctored_exam_id = self._create_proctored_exam()
        self.practice_exam_id = self._create_practice_exam()
        self.onboarding_exam_id = self._create_onboarding_exam()
        self.url = reverse(
            'edx_proctoring:proctored_exam.exam_attempts',
            kwargs={
                'course_id': self.course_id,
            }
        ) + '?' + urlencode({
            'content_id': self.content_id,
            'is_learning_mfe': True,
        })
        self.expected_exam_url = f'{settings.LEARNING_MICROFRONTEND_URL}/course/{self.course_id}/{self.content_id}'
        yesterday = timezone.now() - timezone.timedelta(days=1)
        self.course_scheduled_sections = {
            BlockUsageLocator.from_string(self.content_id_onboarding): MockScheduleItemData(yesterday),
            BlockUsageLocator.from_string(
                'block-v1:test+course+1+type@sequential+block@assignment'
            ): MockScheduleItemData(yesterday),
        }

        set_runtime_service('learning_sequences', MockLearningSequencesService(
            list(self.course_scheduled_sections.keys()),
            self.course_scheduled_sections,
        ))

    def assertHasExamData(self, response_data, has_attempt, has_download_url=False, content_id=None):
        """ Ensure expected exam data is present. """
        exam_data = response_data['exam']
        assert 'exam' in response_data
        assert 'attempt' in exam_data
        if has_attempt:
            assert exam_data['attempt']
            if has_download_url:
                assert 'software_download_url' in exam_data['attempt']
            else:
                assert 'software_download_url' not in exam_data['attempt']
        else:
            assert not exam_data['attempt']
        self.assertEqual(exam_data['course_id'], self.course_id)
        self.assertEqual(exam_data['content_id'], self.content_id if not content_id else content_id)
        self.assertEqual(exam_data['time_limit_mins'], self.default_time_limit)

    def test_staff_get_user_exam_data(self):
        """
        Test staff can get any user's exam data.
        """
        started_attempt = self._create_started_exam_attempt(is_proctored=True)

        # login as different staff user
        admin_user = User(username='test_staff', email='staff@test.com', is_staff=True)
        admin_user.save()
        self.client.login_user(admin_user)
        response = self.client.get(self.url + f'&user_id={self.user.id}')
        response_data = json.loads(response.content.decode('utf-8'))

        self.assertEqual(response_data['active_attempt']['attempt_id'], started_attempt.id)

    def test_non_staff_cannot_get_specific_user_data(self):
        """
        Test non staff cannot get other user's exam data.
        """
        self._create_started_exam_attempt(is_proctored=True)

        other_user = User(username='nefarious_bob', email='tester_bob@test.com')
        other_user.save()
        self.client.login_user(other_user)
        response = self.client.get(self.url + f'&user_id={self.user.id}')
        self.assertEqual(response.status_code, 403)

    def test_exam_total_time_with_allowance_time_before_exam_starts(self):
        """
        Tests that exam has correct total time when user has additional
        time allowance and exam has not started yet.
        """
        allowed_extra_time = 10
        ProctoredExamStudentAllowance.objects.create(
            proctored_exam_id=self.proctored_exam_id,
            user_id=self.user_id,
            key=self.key,
            value=str(allowed_extra_time)
        )
        response = self.client.get(self.url)
        exam = get_exam_by_id(self.proctored_exam_id)
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content.decode('utf-8'))
        expected_total_time = humanized_time(exam['time_limit_mins'] + allowed_extra_time)
        self.assertEqual(response_data['exam']['total_time'], expected_total_time)

    def test_get_started_proctored_exam_attempts_data(self):
        """
        Tests the get proctored exam attempts data by course id and usage key endpoint for started exam.
        """
        self._create_started_exam_attempt(is_proctored=True)

        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content.decode('utf-8'))
        exam_data = response_data['exam']

        # if exam started the prerequisites are not checked
        assert 'prerequisite_status' not in exam_data
        assert 'active_attempt' in response_data and response_data['active_attempt']
        self.assertHasExamData(response_data, has_attempt=True)
        self.assertEqual(exam_data['attempt']['exam_url_path'], self.expected_exam_url)
        self.assertEqual(exam_data['type'], 'proctored')

    def test_get_exam_attempts_data_by_course_id(self):
        """
        Tests the get exam attempts data by course id.
        """
        self._create_started_exam_attempt(is_proctored=True)

        url = reverse(
            'edx_proctoring:proctored_exam.exam_attempts',
            kwargs={
                'course_id': self.course_id,
            }
        ) + '?' + urlencode({
            'is_learning_mfe': True,
        })

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content.decode('utf-8'))
        exam_data = response_data['exam']

        # if exam started the prerequisites are not checked
        assert 'prerequisite_status' not in exam_data
        assert not exam_data
        assert 'active_attempt' in response_data and response_data['active_attempt']

    def test_get_started_timed_exam_attempts_data(self):
        """
        Tests the get timed exam attempts data by course id and usage key endpoint for started exam.
        """
        self._create_started_exam_attempt(is_proctored=False)

        url = reverse(
            'edx_proctoring:proctored_exam.exam_attempts',
            kwargs={
                'course_id': self.course_id,
            }
        ) + '?' + urlencode({
            'content_id': self.content_id_timed,
            'is_learning_mfe': True,
        })

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content.decode('utf-8'))
        exam_data = response_data['exam']
        expected_exam_url = f'{settings.LEARNING_MICROFRONTEND_URL}/course/{self.course_id}/{self.content_id_timed}'
        assert 'prerequisite_status' not in exam_data
        assert 'active_attempt' in response_data and response_data['active_attempt']
        self.assertHasExamData(response_data, has_attempt=True, content_id=self.content_id_timed)
        self.assertEqual(exam_data['attempt']['exam_url_path'], expected_exam_url)

    def test_no_attempts_data_before_exam_starts(self):
        """
        Test we get exam data before exam is started. Ensure no attempts data returned and prerequisites are checked.
        """
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content.decode('utf-8'))
        exam_data = response_data['exam']
        assert 'active_attempt' in response_data and not response_data['active_attempt']
        assert 'prerequisite_status' in exam_data
        self.assertHasExamData(response_data, has_attempt=False)
        self.assertEqual(exam_data['type'], 'proctored')

    def test_get_exam_attempts_data_after_exam_is_submitted(self):
        """
        Tests the Get Exam attempts data contains exam.attempt with submitted status after exam ends.
        """
        self._create_exam_attempt(self.proctored_exam_id, status=ProctoredExamStudentAttemptStatus.submitted)

        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content.decode('utf-8'))
        exam_data = response_data['exam']
        assert 'active_attempt' in response_data and not response_data['active_attempt']
        assert 'prerequisite_status' not in exam_data
        self.assertHasExamData(response_data, has_attempt=True)
        self.assertEqual(exam_data['attempt']['exam_url_path'], self.expected_exam_url)
        self.assertEqual(exam_data['attempt']['attempt_status'], ProctoredExamStudentAttemptStatus.submitted)
        self.assertEqual(exam_data['type'], 'proctored')

    def test_no_exam_data_returned_for_non_exam_sequence(self):
        """
        Test empty exam data is returned for content_id of non-exam sequence item.
        """
        url = reverse(
            'edx_proctoring:proctored_exam.exam_attempts',
            kwargs={
                'course_id': self.course_id,
            }
        ) + '?' + urlencode({
            'content_id': 'block-v1:test+course+1+type@sequential+block@unit',
            'is_learning_mfe': True,
        })
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content.decode('utf-8'))
        exam_data = response_data['exam']
        assert 'active_attempt' in response_data
        assert not response_data['active_attempt']
        assert not exam_data

    @ddt.data(
        ('content_id', True),
        ('content_id_timed', False),
        ('content_id_onboarding', False),
        ('content_id_practice', False),
    )
    @ddt.unpack
    def test_prerequisites_are_not_checked_if_exam_is_not_proctored(self, content_id, should_check_prerequisites):
        """
        Tests that prerequisites are not checked for non proctored exams.
        """
        content_id = getattr(self, content_id)
        url = reverse(
            'edx_proctoring:proctored_exam.exam_attempts',
            kwargs={
                'course_id': self.course_id,
            }
        ) + '?' + urlencode({
            'content_id': content_id,
            'is_learning_mfe': True,
        })

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content.decode('utf-8'))
        exam_data = response_data['exam']
        self.assertEqual('prerequisite_status' in exam_data, should_check_prerequisites)

    @ddt.data(
        ProctoredExamStudentAttemptStatus.created,
        ProctoredExamStudentAttemptStatus.download_software_clicked,
        ProctoredExamStudentAttemptStatus.ready_to_start,
        ProctoredExamStudentAttemptStatus.started,
        ProctoredExamStudentAttemptStatus.ready_to_submit,
        ProctoredExamStudentAttemptStatus.declined,
        ProctoredExamStudentAttemptStatus.submitted,
        ProctoredExamStudentAttemptStatus.rejected,
        ProctoredExamStudentAttemptStatus.expired
    )
    def test_exam_data_contains_necessary_data_based_on_the_attempt_status(self, status):
        """
        Tests the GET exam attempts data contains software download url ONLY when attempt
        is in created or download_software_clicked status
        """
        self._create_exam_attempt(self.proctored_exam_id, status=status)

        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        has_download_url = status in (
            ProctoredExamStudentAttemptStatus.created,
            ProctoredExamStudentAttemptStatus.download_software_clicked
        )
        response_data = json.loads(response.content.decode('utf-8'))
        self.assertHasExamData(
            response_data,
            has_attempt=True,
            has_download_url=has_download_url,
        )

    @ddt.data(
        *product(
            [True, False],
            ProctoredExamStudentAttemptStatus.onboarding_errors
        )
    )
    @ddt.unpack
    def test_exam_data_contains_link_to_onboarding_exam_if_attempt_in_onboarding_errors(self, is_learning_mfe, status):
        """
        Tests the GET exam attempts data contains url to onboarding exam when user
        tries to take proctored exam and has not yet completed required onboarding exam.
        """
        self._create_exam_attempt(self.proctored_exam_id, status=status)

        onboarding_exam = ProctoredExam.objects.get(id=self.onboarding_exam_id)

        if is_learning_mfe:
            expected_exam_url = (
                f'{settings.LEARNING_MICROFRONTEND_URL}/course/'
                f'{self.course_id}/{onboarding_exam.content_id}'
            )
        else:
            expected_exam_url = reverse('jump_to', args=[self.course_id, onboarding_exam.content_id])

        url = reverse(
            'edx_proctoring:proctored_exam.exam_attempts',
            kwargs={
                'course_id': self.course_id,
            }
        ) + '?' + urlencode({
            'content_id': self.content_id,
            'is_learning_mfe': is_learning_mfe,
        })
        response = self.client.get(url)
        response_data = json.loads(response.content.decode('utf-8'))
        exam = response_data['exam']
        assert 'onboarding_link' in exam
        self.assertEqual(expected_exam_url, exam['onboarding_link'])

    def test_exam_data_does_not_fail_if_onboarding_errors_and_no_onboarding_exam(self):
        """
        Tests the GET exam attempts data not contain link to onboarding exam if
        when user tries to take proctored exam and has not yet completed required
        onboarding exam and onboarding exam is not found.
        """
        self._create_exam_attempt(self.proctored_exam_id, status=ProctoredExamStudentAttemptStatus.onboarding_missing)
        url = reverse(
            'edx_proctoring:proctored_exam.exam_attempts',
            kwargs={
                'course_id': self.course_id,
            }
        ) + '?' + urlencode({
            'content_id': self.content_id,
            'is_learning_mfe': True,
        })
        with patch('edx_proctoring.models.ProctoredExam.objects.filter', return_value=ProctoredExam.objects.none()):
            response = self.client.get(url)
        response_data = json.loads(response.content.decode('utf-8'))
        exam = response_data['exam']
        assert 'onboarding_link' in exam
        self.assertEqual(exam['onboarding_link'], '')

    def test_onboarding_errors_and_onboarding_exam_not_available(self):
        """
        Tests the GET exam attempts data not contain link to onboarding exam if
        when user tries to take proctored exam and has not yet completed required
        onboarding exam and onboarding exams are not available for the user.
        """
        self._create_exam_attempt(self.proctored_exam_id, status=ProctoredExamStudentAttemptStatus.onboarding_missing)
        set_runtime_service('learning_sequences', MockLearningSequencesService(
            [],                              # sections user can see (none)
            self.course_scheduled_sections,  # all scheduled sections
        ))
        url = reverse(
            'edx_proctoring:proctored_exam.exam_attempts',
            kwargs={
                'course_id': self.course_id,
            }
        ) + '?' + urlencode({
            'content_id': self.content_id,
            'is_learning_mfe': True,
        })
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content.decode('utf-8'))
        exam = response_data['exam']
        assert 'onboarding_link' in exam
        self.assertEqual(exam['onboarding_link'], '')


# @override_settings(LEARNING_MICROFRONTEND_URL='https//learningmfe', ACCOUNT_MICROFRONTEND_URL='https//localhost')
class ProctoredExamActiveAttemptsMFEViewTests(ProctoredExamTestCase):
    """
    Tests for the ProctoredExamView called from MFE application.
    """
    def setUp(self):
        """
        Initialize
        """
        super().setUp()
        self.proctored_exam_id = self._create_proctored_exam()
        self.url = reverse('edx_proctoring:proctored_exam.active_attempt')

    def test_get_started_attempt(self):
        """
        Tests the get attempt for user with started attempt.
        """
        started_attempt = self._create_started_exam_attempt(is_proctored=True)
        self._create_exam_attempt(
            self.proctored_exam_id,
            ProctoredExamStudentAttemptStatus.rejected,
        )

        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content.decode('utf-8'))

        self.assertEqual(response_data['attempt_id'], started_attempt.id)

    def test_get_no_started_attempts(self):
        """
        Tests the get attempt for user with no started attempts.
        """
        self._create_exam_attempt(
            self.proctored_exam_id,
            ProctoredExamStudentAttemptStatus.submitted,
        )

        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content.decode('utf-8'))

        self.assertEqual(response_data, {})

    def test_staff_get_user_attempt(self):
        """
        Test staff can get any user's started attempt.
        """
        started_attempt = self._create_started_exam_attempt(is_proctored=True)

        # login as different staff user
        admin_user = User(username='test_staff', email='staff@test.com', is_staff=True)
        admin_user.save()
        self.client.login_user(admin_user)
        response = self.client.get(self.url + f'?user_id={self.user.id}')
        response_data = json.loads(response.content.decode('utf-8'))

        self.assertEqual(response_data['attempt_id'], started_attempt.id)

    def test_non_staff_cannot_get_user_attempt(self):
        """
        Test non staff cannot get another user's attempt.
        """
        self._create_started_exam_attempt(is_proctored=True)

        other_user = User(username='nefarious_bob', email='tester_bob@test.com')
        other_user.save()
        self.client.login_user(other_user)
        response = self.client.get(self.url + f'?user_id={self.user.id}')
        self.assertEqual(response.status_code, 403)


class ProctoredSettingsViewTests(ProctoredExamTestCase):
    """
    Tests for the ProctoredSettingsView.
    """

    def setUp(self):
        """
        Initialize.
        """
        super().setUp()
        self.proctored_exam_id = self._create_proctored_exam()
        self.timed_exam_id = self._create_timed_exam()

    def get_url(self, exam_id):
        """
        Returns ProctoredSettingsView url for exam_id.
        """
        return reverse(
            'edx_proctoring:proctored_exam.proctoring_settings',
            kwargs={
                'exam_id': exam_id,
            }
        )

    def test_get_proctoring_settings_for_proctored_exam(self):
        """
        Tests the get proctoring settings for proctored exam.
        """
        url = self.get_url(self.proctored_exam_id)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content.decode('utf-8'))
        assert 'link_urls' in response_data and response_data['link_urls']
        assert 'exam_proctoring_backend' in response_data

    def test_get_proctoring_settings_for_timed_exam(self):
        """
        Tests the get proctoring settings for timed exam.
        """
        url = self.get_url(self.timed_exam_id)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content.decode('utf-8'))
        assert 'link_urls' in response_data and response_data['link_urls']
        assert 'exam_proctoring_backend' in response_data and not response_data['exam_proctoring_backend']

    def test_get_proctoring_settings_for_non_existing_exam(self):
        """
        Test to get the proctoring settings for non-existing exam_id raises exception.
        """
        exam_id = 9999
        url = self.get_url(exam_id)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 400)
        self.assertRaises(ProctoredExamNotFoundException)

    def test_get_proctoring_settings_for_exam_with_not_configured_backend(self):
        """
        Test to get the proctoring settings for the exam with not configured backend raises an exception.
        """
        url = self.get_url(self.proctored_exam_id)
        with patch('edx_proctoring.api.get_backend_provider', side_effect=NotImplementedError()):
            response = self.client.get(url)
        self.assertEqual(response.status_code, 500)
        self.assertRaises(BackendProviderNotConfigured)


class ProctoredExamReviewPolicyView(ProctoredExamTestCase):
    """
    Tests for the ProctoredExamReviewPolicyView.
    """

    def setUp(self):
        """
        Initialize.
        """
        super().setUp()
        self.proctored_exam_id = self._create_proctored_exam()
        self.review_policy_id = self._create_review_policy(self.proctored_exam_id)
        self.url = reverse(
            'edx_proctoring:proctored_exam.review_policy',
            kwargs={
                'exam_id': self.proctored_exam_id,
            }
        )

    def test_get_exam_review_policy_for_proctored_exam(self):
        """
        Tests the GET exam review policy endpoint for proctored exam with existing policy.
        """
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content.decode('utf-8'))
        expected_review_policy = get_review_policy_by_exam_id(self.proctored_exam_id)
        assert 'review_policy' in response_data
        self.assertEqual(response_data['review_policy'], expected_review_policy['review_policy'])

    def test_get_exam_review_policy_for_proctored_exam_with_no_existing_review(self):
        """
        Tests the GET exam review policy endpoint for proctored exam which has no review policy configured.
        """
        with patch('edx_proctoring.models.ProctoredExamReviewPolicy.get_review_policy_for_exam', return_value=None):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        response_data = json.loads(response.content.decode('utf-8'))
        assert 'review_policy' in response_data
        self.assertIsNone(response_data['review_policy'])



"""edx-proctoring signals"""
from django.dispatch import Signal

# Signal that is emitted when an attempt status is updated. Added to utils to avoid cyclic import in signals.py file
exam_attempt_status_signal = Signal()



"""
Django management command to update the status of a ProctoredExamStudentAttempt
based on the status of its corresponding ProctoredExamSoftwareSecureReview
"""
import datetime
import logging
import time

from django.core.management.base import BaseCommand

from edx_proctoring.models import ProctoredExamSoftwareSecureReview

log = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Django Management command to update is_attempt_active field on review models
    """

    def add_arguments(self, parser):
        parser.add_argument(
          '--start_date_time',
          action='store',
          dest='start_date_time',
          type=str,
          help='First date time for reviews that we want to consider. Should be formatted as 2020-12-02 00:00:00.'
        )
        parser.add_argument(
          '--end_date_time',
          action='store',
          dest='end_date_time',
          type=str,
          help='Last date time for reviews that we want to consider. Should be formatted as 2020-12-02 00:00:00.'
        )
        parser.add_argument(
            '--batch_size',
            action='store',
            dest='batch_size',
            type=int,
            default=300,
            help='Maximum number of attempt_codes to process. '
                 'This helps avoid overloading the database while updating large amount of data.'
        )
        parser.add_argument(
            '--sleep_time',
            action='store',
            dest='sleep_time',
            type=int,
            default=10,
            help='Sleep time in seconds between update of batches'
        )

    def handle(self, *args, **options):
        """
        Management command entry point, simply call into the signal firing
        """
        batch_size = options['batch_size']
        sleep_time = options['sleep_time']
        review_start_date = datetime.datetime.strptime(options['start_date_time'], '%Y-%m-%d %H:%M:%S')
        review_end_date = datetime.datetime.strptime(options['end_date_time'], '%Y-%m-%d %H:%M:%S')

        reviews_in_date_range = ProctoredExamSoftwareSecureReview.objects.filter(
          modified__range=[review_start_date, review_end_date]
        )
        review_count = 0

        for review in reviews_in_date_range:
            review_id = review.id
            attempt_code = review.attempt_code
            log.info(
                'Saving review_id=%i for corresponding attempt_code=%s',
                review_id,
                attempt_code
            )
            review.save()
            review_count += 1

            if review_count == batch_size:
                review_count = 0
                time.sleep(sleep_time)



"""
URL mappings for edX Proctoring Server.
"""

from django.conf import settings
from django.urls import include, path, re_path

from edx_proctoring import callbacks, instructor_dashboard_exam_urls, views

app_name = 'edx_proctoring'

CONTENT_ID_PATTERN = r'(?P<content_id>([A-z0-9]+|(?:i4x://?[^/]+/[^/]+/[^/]+/[^@]+(?:@[^/]+)?)|(?:[^/]+)))'

urlpatterns = [
    path('edx_proctoring/v1/proctored_exam/exam', views.ProctoredExamView.as_view(),
         name='proctored_exam.exam'
         ),
    path('edx_proctoring/v1/proctored_exam/exam/exam_id/<int:exam_id>', views.ProctoredExamView.as_view(),
         name='proctored_exam.exam_by_id'
         ),
    re_path(
        (fr'edx_proctoring/v1/proctored_exam/exam/course_id/{settings.COURSE_ID_PATTERN}'
         '/content_id/(?P<content_id>[A-z0-9]+)$'),
        views.ProctoredExamView.as_view(),
        name='proctored_exam.exam_by_content_id'
    ),
    re_path(
        fr'edx_proctoring/v1/proctored_exam/exam/course_id/{settings.COURSE_ID_PATTERN}$',
        views.ProctoredExamView.as_view(),
        name='proctored_exam.exams_by_course_id'
    ),
    re_path(
        fr'edx_proctoring/v1/proctored_exam/exam_registration/course_id/{settings.COURSE_ID_PATTERN}$',
        views.RegisterProctoredExamsView.as_view(),
        name='proctored_exam.register_exams_by_course_id'
    ),
    path('edx_proctoring/v1/proctored_exam/attempt/<int:attempt_id>', views.StudentProctoredExamAttempt.as_view(),
         name='proctored_exam.attempt'
         ),
    re_path(
        fr'edx_proctoring/v1/proctored_exam/attempt/grouped/course_id/{settings.COURSE_ID_PATTERN}$',
        views.StudentProctoredGroupedExamAttemptsByCourse.as_view(),
        name='proctored_exam.attempts.grouped.course'
    ),
    re_path(
        'edx_proctoring/v1/proctored_exam/attempt/grouped/course_id/'
        fr'{settings.COURSE_ID_PATTERN}/search/(?P<search_by>.+)$',
        views.StudentProctoredGroupedExamAttemptsByCourse.as_view(),
        name='proctored_exam.attempts.grouped.search'
    ),
    path('edx_proctoring/v1/proctored_exam/attempt',
         views.StudentProctoredExamAttemptCollection.as_view(),
         name='proctored_exam.attempt.collection'
         ),
    path('edx_proctoring/v1/proctored_exam/attempt/<int:attempt_id>/review_status',
         views.ProctoredExamAttemptReviewStatus.as_view(),
         name='proctored_exam.attempt.review_status'
         ),
    path('edx_proctoring/v1/proctored_exam/attempt/<slug:external_id>/ready',
         views.ExamReadyCallback.as_view(),
         name='proctored_exam.attempt.ready_callback'
         ),
    path('edx_proctoring/v1/proctored_exam/attempt/<slug:external_id>/reviewed',
         views.ProctoredExamReviewCallback.as_view(),
         name='proctored_exam.attempt.callback'
         ),
    re_path(
        fr'edx_proctoring/v1/proctored_exam/{settings.COURSE_ID_PATTERN}/allowance$',
        views.ExamAllowanceView.as_view(),
        name='proctored_exam.allowance'
    ),
    path('edx_proctoring/v1/proctored_exam/allowance', views.ExamAllowanceView.as_view(),
         name='proctored_exam.allowance'
         ),
    re_path(
        fr'edx_proctoring/v1/proctored_exam/{settings.COURSE_ID_PATTERN}/bulk_allowance$',
        views.ExamBulkAllowanceView.as_view(),
        name='proctored_exam.bulk_allowance'
    ),
    path('edx_proctoring/v1/proctored_exam/bulk_allowance', views.ExamBulkAllowanceView.as_view(),
         name='proctored_exam.bulk_allowance'
         ),
    re_path(
        fr'edx_proctoring/v1/proctored_exam/{settings.COURSE_ID_PATTERN}/grouped/allowance$',
        views.GroupedExamAllowancesByStudent.as_view(),
        name='proctored_exam.allowance.grouped.course'
    ),
    path('edx_proctoring/v1/proctored_exam/active_exams_for_user', views.ActiveExamsForUserView.as_view(),
         name='proctored_exam.active_exams_for_user'
         ),
    path('edx_proctoring/v1/user_onboarding/status', views.StudentOnboardingStatusView.as_view(),
         name='user_onboarding.status'
         ),
    re_path(
        fr'edx_proctoring/v1/user_onboarding/status/course_id/{settings.COURSE_ID_PATTERN}$',
        views.StudentOnboardingStatusByCourseView.as_view(),
        name='user_onboarding.status.course'
    ),
    re_path(
        fr'edx_proctoring/v1/instructor/{settings.COURSE_ID_PATTERN}$',
        views.InstructorDashboard.as_view(),
        name='instructor_dashboard_course'
    ),
    re_path(
        r'edx_proctoring/v1/retire_backend_user/(?P<user_id>[\d]+)/$',
        views.BackendUserManagementAPI.as_view(),
        name='backend_user_deletion_api'
    ),
    re_path(
        r'edx_proctoring/v1/retire_user/(?P<user_id>[\d]+)/$',
        views.UserRetirement.as_view(),
        name='user_retirement_api'
    ),
    re_path(
        fr'edx_proctoring/v1/proctored_exam/attempt/course_id/{settings.COURSE_ID_PATTERN}$',
        views.ProctoredExamAttemptView.as_view(),
        name='proctored_exam.exam_attempts'
    ),
    re_path(
        'edx_proctoring/v1/proctored_exam/active_attempt',
        views.ProctoredExamActiveAttemptView.as_view(),
        name='proctored_exam.active_attempt'
    ),
    path('edx_proctoring/v1/proctored_exam/settings/exam_id/<int:exam_id>/', views.ProctoredSettingsView.as_view(),
         name='proctored_exam.proctoring_settings'
         ),
    path('edx_proctoring/v1/proctored_exam/review_policy/exam_id/<int:exam_id>/',
         views.ProctoredExamReviewPolicyView.as_view(),
         name='proctored_exam.review_policy'
         ),

    # Unauthenticated callbacks from SoftwareSecure. Note we use other
    # security token measures to protect data
    #
    path('edx_proctoring/proctoring_launch_callback/start_exam/<slug:attempt_code>', callbacks.start_exam_callback,
         name='anonymous.proctoring_launch_callback.start_exam'
         ),
    path('edx_proctoring/proctoring_review_callback/', views.AnonymousReviewCallback.as_view(),
         name='anonymous.proctoring_review_callback'
         ),
    re_path(
        r'edx_proctoring/v1/proctored_exam/exam_id/(?P<exam_id>\d+)/user_id/(?P<user_id>[\d]+)/reset_attempts$',
        views.StudentProctoredExamResetAttempts.as_view(),
        name='proctored_exam.attempts.reset'
    ),
    path('', include('rest_framework.urls', namespace='rest_framework')),
]

urlpatterns += instructor_dashboard_exam_urls.urlpatterns
